
exception Unimplemented
exception EvalFailed
exception BadCast
exception UninitField

type varname   = string
type classname = string
type fieldname = string
type methname = string

type exp = 
  | Self
  | New of classname
  | Call of exp * methname * (exp list)
  | Getfield of exp * fieldname 
  | Setfield of exp * fieldname * exp
  | Localvar of varname
  | LetIn of varname * classname * exp * exp
      (* let x : t = e1 in e2 -- can also be used for sequencing *)
  | Cast of classname * exp
  | Fail

(* (return type, argument names and types, body) *)
type meth = classname * ((classname * varname) list) * exp

(* 
  * 'a is classname option before translation (the immediate superclass)
    and classname list after translation (all proper superclasses)
  * methods is _new and overridden_ methods before translation and _all_
    methods after translation
  * fields is _new_ fields before translation and _all_ fields after translation
*)
type 'a classdef = {
    myname    : classname;
    super     : 'a; 
    methods   : (methname * meth) list;
    fields    : (fieldname * classname) list
  }

(* 'a is as for classdef *)
type 'a program = ('a classdef list) * exp

let objectclass = {
    myname  = "object";
    super   = None;
    methods = [];
    fields  = []
} 

let boolclass = {
  myname  = "bool";
  super   = Some "object";
  methods = [("ifThenElse", ("object", [("object","x");("object","y")], 
			     Fail))];
  fields  = []
} 

(******* part (a) **********)

let trueclass = raise Unimplemented 
let falseclass = raise Unimplemented 

(******* part (b) **********)

let translate_classdefs classdefs = raise Unimplemented

(******* part (c) **********)

type valu = classname * ((fieldname * (field_contents ref)) list)
and field_contents = Object of valu | Uninit

let rec interp classdefs env self exp =
  let getvar   v = List.assoc v env in
  let getclass c = List.find (fun cd -> cd.myname=c) classdefs in
  match exp with
    Fail -> raise EvalFailed
  | Self -> self
  | New c -> 
      let fields = List.map (fun (x,_) -> (x,ref Uninit)) (getclass c).fields in
      (c, fields)
  | _ -> raise Unimplemented

let run_it (classdefs,main) =
  interp (translate_classdefs classdefs) [] ("object",[]) main

